# encoding: utf-8

# This file is part of CycloneDX Python Lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

import io
import json
import os
import sys
import xml.etree.ElementTree
from datetime import datetime, timezone
from typing import Any
from unittest import TestCase
from uuid import uuid4

from lxml import etree
from lxml.etree import DocumentInvalid
from xmldiff import main
from xmldiff.actions import MoveNode

from cyclonedx.output import SchemaVersion

if sys.version_info >= (3, 7):
    from jsonschema import ValidationError, validate as json_validate

if sys.version_info >= (3, 8):
    from importlib.metadata import version
else:
    from importlib_metadata import version

cyclonedx_lib_name: str = 'cyclonedx-python-lib'
cyclonedx_lib_version: str = version(cyclonedx_lib_name)
single_uuid: str = 'urn:uuid:{}'.format(uuid4())
schema_directory = os.path.join(os.path.dirname(__file__), '..', 'cyclonedx', 'schema')


class BaseJsonTestCase(TestCase):

    def assertValidAgainstSchema(self, bom_json: str, schema_version: SchemaVersion) -> None:
        if sys.version_info >= (3, 7):
            schema_fn = os.path.join(
                schema_directory,
                f'bom-{schema_version.name.replace("_", ".").replace("V", "")}.schema.json'
            )
            with open(schema_fn) as schema_fd:
                schema_doc = json.load(schema_fd)

            try:
                json_validate(instance=json.loads(bom_json), schema=schema_doc)
            except ValidationError as e:
                self.assertTrue(False, f'Failed to validate SBOM against JSON schema: {str(e)}')

            self.assertTrue(True)
        else:
            self.assertTrue(True, 'JSON Schema Validation is not possible in Python < 3.7')

    @staticmethod
    def _sort_json_dict(item: object) -> Any:
        if isinstance(item, dict):
            return sorted((key, BaseJsonTestCase._sort_json_dict(values)) for key, values in item.items())
        if isinstance(item, list):
            return sorted(BaseJsonTestCase._sort_json_dict(x) for x in item)
        else:
            return item

    def assertEqualJson(self, a: str, b: str) -> None:
        self.assertEqual(
            BaseJsonTestCase._sort_json_dict(json.loads(a)),
            BaseJsonTestCase._sort_json_dict(json.loads(b))
        )

    def assertEqualJsonBom(self, a: str, b: str) -> None:
        """
        Remove UUID before comparison as this will be unique to each generation
        """
        ab = json.loads(a)
        bb = json.loads(b)

        # Null serialNumbers
        ab['serialNumber'] = single_uuid
        bb['serialNumber'] = single_uuid

        # Unify timestamps to ensure they will compare
        now = datetime.now(tz=timezone.utc)
        if 'metadata' in ab.keys():
            ab['metadata']['timestamp'] = now.isoformat()
        if 'metadata' in bb.keys():
            bb['metadata']['timestamp'] = now.isoformat()

        self.assertEqualJson(json.dumps(ab), json.dumps(bb))


class BaseXmlTestCase(TestCase):

    def assertValidAgainstSchema(self, bom_xml: str, schema_version: SchemaVersion) -> None:
        xsd_fn = os.path.join(schema_directory, f'bom-{schema_version.name.replace("_", ".").replace("V", "")}.xsd')
        with open(xsd_fn) as xsd_fd:
            xsd_doc = etree.parse(xsd_fd)

        xml_schema = etree.XMLSchema(xsd_doc)
        schema_validates = False
        try:
            schema_validates = xml_schema.validate(etree.parse(io.BytesIO(bytes(bom_xml, 'ascii'))))
        except DocumentInvalid as e:
            print(f'Failed to validate SBOM against schema: {str(e)}')

        if not schema_validates:
            print(xml_schema.error_log.last_error)
        self.assertTrue(schema_validates, f'Failed to validate Generated SBOM against XSD Schema:'
                                          f'{bom_xml}')

    def assertEqualXml(self, a: str, b: str) -> None:
        diff_results = main.diff_texts(a, b, diff_options={'F': 0.5})
        diff_results = list(filter(lambda o: not isinstance(o, MoveNode), diff_results))
        self.assertEqual(len(diff_results), 0, f'There are XML differences: {diff_results}\n- {a}\n+ {b}')

    def assertEqualXmlBom(self, a: str, b: str, namespace: str) -> None:
        """
        Sanitise some fields such as timestamps which cannot have their values directly compared for equality.
        """
        ba = xml.etree.ElementTree.fromstring(a, etree.XMLParser(remove_blank_text=True, remove_comments=True))
        bb = xml.etree.ElementTree.fromstring(b, etree.XMLParser(remove_blank_text=True, remove_comments=True))

        # Align serialNumbers
        ba.set('serialNumber', single_uuid)
        bb.set('serialNumber', single_uuid)

        # Align timestamps in metadata
        now = datetime.now(tz=timezone.utc)
        metadata_ts_a = ba.find('./{{{}}}metadata/{{{}}}timestamp'.format(namespace, namespace))
        if metadata_ts_a is not None:
            metadata_ts_a.text = now.isoformat()
        metadata_ts_b = bb.find('./{{{}}}metadata/{{{}}}timestamp'.format(namespace, namespace))
        if metadata_ts_b is not None:
            metadata_ts_b.text = now.isoformat()

        self.assertEqualXml(
            xml.etree.ElementTree.tostring(ba, 'unicode'),
            xml.etree.ElementTree.tostring(bb, 'unicode')
        )



# encoding: utf-8

# This file is part of CycloneDX Python Lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

from enum import Enum

"""
This set of classes represents the data about Impact Analysis.

Impact Analysis is new for CycloneDX schema version 1.

.. note::
    See the CycloneDX Schema extension definition https://cyclonedx.org/docs/1.4
"""


class ImpactAnalysisAffectedStatus(str, Enum):
    """
    Enum object that defines the permissible impact analysis affected states.

    The vulnerability status of a given version or range of versions of a product.

    The statuses 'affected' and 'unaffected' indicate that the version is affected or unaffected by the vulnerability.

    The status 'unknown' indicates that it is unknown or unspecified whether the given version is affected. There can
    be many reasons for an 'unknown' status, including that an investigation has not been undertaken or that a vendor
    has not disclosed the status.

    .. note::
        See the CycloneDX Schema definition: https://cyclonedx.org/docs/1.4/#type_impactAnalysisAffectedStatusType
    """

    AFFECTED = 'affected'
    UNAFFECTED = 'unaffected'
    UNKNOWN = 'unknown'


class ImpactAnalysisJustification(str, Enum):
    """
    Enum object that defines the rationale of why the impact analysis state was asserted.

    .. note::
        See the CycloneDX Schema definition: https://cyclonedx.org/docs/1.4/#type_impactAnalysisJustificationType
    """

    CODE_NOT_PRESENT = 'code_not_present'
    CODE_NOT_REACHABLE = 'code_not_reachable'
    PROTECTED_AT_PERIMITER = 'protected_at_perimeter'
    PROTECTED_AT_RUNTIME = 'protected_at_runtime'
    PROTECTED_BY_COMPILER = 'protected_by_compiler'
    PROTECTED_BY_MITIGATING_CONTROL = 'protected_by_mitigating_control'
    REQUIRES_CONFIGURATION = 'requires_configuration'
    REQUIRES_DEPENDENCY = 'requires_dependency'
    REQUIRES_ENVIRONMENT = 'requires_environment'


class ImpactAnalysisResponse(str, Enum):
    """
    Enum object that defines the valid rationales as to why the impact analysis state was asserted.

    .. note::
        See the CycloneDX Schema definition: https://cyclonedx.org/docs/1.4/#type_impactAnalysisResponsesType
    """

    CAN_NOT_FIX = 'can_not_fix'
    ROLLBACK = 'rollback'
    UPDATE = 'update'
    WILL_NOT_FIX = 'will_not_fix'
    WORKAROUND_AVAILABLE = 'workaround_available'


class ImpactAnalysisState(str, Enum):
    """
    Enum object that defines the permissible impact analysis states.

    .. note::
        See the CycloneDX Schema definition: https://cyclonedx.org/docs/1.4/#type_impactAnalysisStateType
    """

    RESOLVED = 'resolved'
    RESOLVED_WITH_PEDIGREE = 'resolved_with_pedigree'
    EXPLOITABLE = 'exploitable'
    IN_TRIAGE = 'in_triage'
    FALSE_POSITIVE = 'false_positive'
    NOT_AFFECTED = 'not_affected'



# encoding: utf-8

# This file is part of CycloneDX Python Lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.
import warnings
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Iterable, Optional, Set
from uuid import UUID, uuid4

from sortedcontainers import SortedSet

from ..exception.model import UnknownComponentDependencyException
from ..parser import BaseParser
from . import ExternalReference, LicenseChoice, OrganizationalContact, OrganizationalEntity, Property, ThisTool, Tool
from .component import Component
from .service import Service

if TYPE_CHECKING:
    from packageurl import PackageURL  # type:ignore[import]


class BomMetaData:
    """
    This is our internal representation of the metadata complex type within the CycloneDX standard.

    .. note::
        See the CycloneDX Schema for Bom metadata: https://cyclonedx.org/docs/1.4/#type_metadata
    """

    def __init__(self, *, tools: Optional[Iterable[Tool]] = None,
                 authors: Optional[Iterable[OrganizationalContact]] = None, component: Optional[Component] = None,
                 manufacture: Optional[OrganizationalEntity] = None,
                 supplier: Optional[OrganizationalEntity] = None,
                 licenses: Optional[Iterable[LicenseChoice]] = None,
                 properties: Optional[Iterable[Property]] = None) -> None:
        self.timestamp = datetime.now(tz=timezone.utc)
        self.tools = tools or []  # type: ignore
        self.authors = authors or []  # type: ignore
        self.component = component
        self.manufacture = manufacture
        self.supplier = supplier
        self.licenses = licenses or []  # type: ignore
        self.properties = properties or []  # type: ignore

        if not tools:
            self.tools.add(ThisTool)

    @property
    def timestamp(self) -> datetime:
        """
        The date and time (in UTC) when this BomMetaData was created.

        Returns:
            `datetime` instance in UTC timezone
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: datetime) -> None:
        self._timestamp = timestamp

    @property
    def tools(self) -> "SortedSet[Tool]":
        """
        Tools used to create this BOM.

        Returns:
            `Set` of `Tool` objects.
        """
        return self._tools

    @tools.setter
    def tools(self, tools: Iterable[Tool]) -> None:
        self._tools = SortedSet(tools)

    @property
    def authors(self) -> "SortedSet[OrganizationalContact]":
        """
        The person(s) who created the BOM.

        Authors are common in BOMs created through manual processes.

        BOMs created through automated means may not have authors.

        Returns:
            Set of `OrganizationalContact`
        """
        return self._authors

    @authors.setter
    def authors(self, authors: Iterable[OrganizationalContact]) -> None:
        self._authors = SortedSet(authors)

    @property
    def component(self) -> Optional[Component]:
        """
        The (optional) component that the BOM describes.

        Returns:
            `cyclonedx.model.component.Component` instance for this Bom Metadata.
        """
        return self._component

    @component.setter
    def component(self, component: Component) -> None:
        """
        The (optional) component that the BOM describes.

        Args:
            component
                `cyclonedx.model.component.Component` instance to add to this Bom Metadata.

        Returns:
            None
        """
        self._component = component

    @property
    def manufacture(self) -> Optional[OrganizationalEntity]:
        """
        The organization that manufactured the component that the BOM describes.

        Returns:
            `OrganizationalEntity` if set else `None`
        """
        return self._manufacture

    @manufacture.setter
    def manufacture(self, manufacture: Optional[OrganizationalEntity]) -> None:
        self._manufacture = manufacture

    @property
    def supplier(self) -> Optional[OrganizationalEntity]:
        """
        The organization that supplied the component that the BOM describes.

        The supplier may often be the manufacturer, but may also be a distributor or repackager.

        Returns:
            `OrganizationalEntity` if set else `None`
        """
        return self._supplier

    @supplier.setter
    def supplier(self, supplier: Optional[OrganizationalEntity]) -> None:
        self._supplier = supplier

    @property
    def licenses(self) -> "SortedSet[LicenseChoice]":
        """
        A optional list of statements about how this BOM is licensed.

        Returns:
            Set of `LicenseChoice`
        """
        return self._licenses

    @licenses.setter
    def licenses(self, licenses: Iterable[LicenseChoice]) -> None:
        self._licenses = SortedSet(licenses)

    @property
    def properties(self) -> "SortedSet[Property]":
        """
        Provides the ability to document properties in a key/value store. This provides flexibility to include data not
        officially supported in the standard without having to use additional namespaces or create extensions.

        Property names of interest to the general public are encouraged to be registered in the CycloneDX Property
        Taxonomy - https://github.com/CycloneDX/cyclonedx-property-taxonomy. Formal registration is OPTIONAL.

        Return:
            Set of `Property`
        """
        return self._properties

    @properties.setter
    def properties(self, properties: Iterable[Property]) -> None:
        self._properties = SortedSet(properties)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BomMetaData):
            return hash(other) == hash(self)
        return False

    def __hash__(self) -> int:
        return hash((
            self.timestamp, self.tools, self.component
        ))

    def __repr__(self) -> str:
        return f'<BomMetaData timestamp={self.timestamp.utcnow()}>'


class Bom:
    """
    This is our internal representation of a bill-of-materials (BOM).

    You can either create a `cyclonedx.model.bom.Bom` yourself programmatically, or generate a `cyclonedx.model.bom.Bom`
    from a `cyclonedx.parser.BaseParser` implementation.

    Once you have an instance of `cyclonedx.model.bom.Bom`, you can pass this to an instance of
    `cyclonedx.output.BaseOutput` to produce a CycloneDX document according to a specific schema version and format.
    """

    @staticmethod
    def from_parser(parser: BaseParser) -> 'Bom':
        """
        Create a Bom instance from a Parser object.

        Args:
            parser (`cyclonedx.parser.BaseParser`): A valid parser instance.

        Returns:
            `cyclonedx.model.bom.Bom`: A Bom instance that represents the valid data held in the supplied parser.
        """
        bom = Bom()
        bom.components.update(parser.get_components())
        return bom

    def __init__(self, *, components: Optional[Iterable[Component]] = None,
                 services: Optional[Iterable[Service]] = None,
                 external_references: Optional[Iterable[ExternalReference]] = None) -> None:
        """
        Create a new Bom that you can manually/programmatically add data to later.

        Returns:
            New, empty `cyclonedx.model.bom.Bom` instance.
        """
        self.uuid = uuid4()
        self.metadata = BomMetaData()
        self.components = components or []  # type: ignore
        self.services = services or []  # type: ignore
        self.external_references = external_references or []  # type: ignore

    @property
    def uuid(self) -> UUID:
        """
        Unique UUID for this BOM

        Returns:
            `UUID` instance
        """
        return self.__uuid

    @uuid.setter
    def uuid(self, uuid: UUID) -> None:
        self.__uuid = uuid

    @property
    def metadata(self) -> BomMetaData:
        """
        Get our internal metadata object for this Bom.

        Returns:
            Metadata object instance for this Bom.

        .. note::
            See the CycloneDX Schema for Bom metadata: https://cyclonedx.org/docs/1.3/#type_metadata
        """
        return self._metadata

    @metadata.setter
    def metadata(self, metadata: BomMetaData) -> None:
        self._metadata = metadata

    @property
    def components(self) -> "SortedSet[Component]":
        """
        Get all the Components currently in this Bom.

        Returns:
             Set of `Component` in this Bom
        """
        return self._components

    @components.setter
    def components(self, components: Iterable[Component]) -> None:
        self._components = SortedSet(components)

    def get_component_by_purl(self, purl: Optional["PackageURL"]) -> Optional[Component]:
        """
        Get a Component already in the Bom by its PURL

        Args:
             purl:
                An instance of `packageurl.PackageURL` to look and find `Component`.

        Returns:
            `Component` or `None`
        """
        if purl:
            found = [x for x in self.components if x.purl == purl]
            if len(found) == 1:
                return found[0]

        return None

    def get_urn_uuid(self) -> str:
        """
        Get the unique reference for this Bom.

        Returns:
            URN formatted UUID that uniquely identified this Bom instance.
        """
        return 'urn:uuid:{}'.format(self.__uuid)

    def has_component(self, component: Component) -> bool:
        """
        Check whether this Bom contains the provided Component.

        Args:
            component:
                The instance of `cyclonedx.model.component.Component` to check if this Bom contains.

        Returns:
            `bool` - `True` if the supplied Component is part of this Bom, `False` otherwise.
        """
        return component in self.components

    @property
    def services(self) -> "SortedSet[Service]":
        """
        Get all the Services currently in this Bom.

        Returns:
             Set of `Service` in this BOM
        """
        return self._services

    @services.setter
    def services(self, services: Iterable[Service]) -> None:
        self._services = SortedSet(services)

    @property
    def external_references(self) -> "SortedSet[ExternalReference]":
        """
        Provides the ability to document external references related to the BOM or to the project the BOM describes.

        Returns:
            Set of `ExternalReference`
        """
        return self._external_references

    @external_references.setter
    def external_references(self, external_references: Iterable[ExternalReference]) -> None:
        self._external_references = SortedSet(external_references)

    def _get_all_components(self) -> Set[Component]:
        components: Set[Component] = set()
        if self.metadata.component:
            components.update(self.metadata.component.get_all_nested_components(include_self=True))

        for c in self.components:
            components.update(c.get_all_nested_components(include_self=True))

        return components

    def has_vulnerabilities(self) -> bool:
        """
        Check whether this Bom has any declared vulnerabilities.

        Returns:
            `bool` - `True` if at least one `cyclonedx.model.component.Component` has at least one Vulnerability,
                `False` otherwise.
        """
        return any(c.has_vulnerabilities() for c in self.components)

    def validate(self) -> bool:
        """
        Perform data-model level validations to make sure we have some known data integrity prior to attempting output
        of this `Bom`

        Returns:
             `bool`
        """

        # 1. Make sure dependencies are all in this Bom.
        all_bom_refs = set(map(lambda c: c.bom_ref, self._get_all_components())) | set(
            map(lambda s: s.bom_ref, self.services))

        all_dependency_bom_refs = set().union(*(c.dependencies for c in self.components))
        dependency_diff = all_dependency_bom_refs - all_bom_refs
        if len(dependency_diff) > 0:
            raise UnknownComponentDependencyException(
                f'One or more Components have Dependency references to Components/Services that are not known in this '
                f'BOM. They are: {dependency_diff}')

        # 2. Dependencies should exist for the Component this BOM is describing, if one is set
        if self.metadata.component and not self.metadata.component.dependencies:
            warnings.warn(
                f'The Component this BOM is describing (PURL={self.metadata.component.purl}) has no defined '
                f'dependencies which means the Dependency Graph is incomplete - you should add direct dependencies to '
                f'this Component to complete the Dependency Graph data.',
                UserWarning
            )

        return True

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Bom):
            return hash(other) == hash(self)
        return False

    def __hash__(self) -> int:
        return hash((
            self.uuid, self.metadata, tuple(self.components), tuple(self.services), tuple(self.external_references)
        ))

    def __repr__(self) -> str:
        return f'<Bom uuid={self.uuid}>'



# encoding: utf-8

# This file is part of CycloneDX Python Lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

from typing import Iterable, Optional

from sortedcontainers import SortedSet

from .bom_ref import BomRef


class Dependency:
    """
    This is our internal representation of a Dependency for a Component.

    .. note::
        See the CycloneDX Schema definition: https://cyclonedx.org/docs/1.4/xml/#type_dependencyType
    """

    def __init__(self, *, ref: BomRef, depends_on: Optional[Iterable[BomRef]] = None) -> None:
        self._ref = ref
        self.depends_on = depends_on or []  # type: ignore

    @property
    def ref(self) -> BomRef:
        return self._ref

    @property
    def depends_on(self) -> "SortedSet[BomRef]":
        return self._depends_on

    @depends_on.setter
    def depends_on(self, depends_on: Iterable[BomRef]) -> None:
        self._depends_on = SortedSet(depends_on)



# encoding: utf-8

# This file is part of CycloneDX Python Lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

from os.path import join
from unittest.mock import Mock, patch

from cyclonedx.exception.model import UnknownComponentDependencyException
from cyclonedx.model.bom import Bom
from cyclonedx.output import SchemaVersion, get_instance

from . import FIXTURES_DIRECTORY, RECREATE_SNAPSHOTS
from .base import BaseXmlTestCase
from .data import (
    MOCK_UUID_1,
    MOCK_UUID_2,
    MOCK_UUID_3,
    MOCK_UUID_4,
    MOCK_UUID_5,
    MOCK_UUID_6,
    TEST_UUIDS,
    get_bom_for_issue_275_components,
    get_bom_for_issue_328_components,
    get_bom_just_complete_metadata,
    get_bom_with_component_setuptools_basic,
    get_bom_with_component_setuptools_complete,
    get_bom_with_component_setuptools_no_component_version,
    get_bom_with_component_setuptools_with_cpe,
    get_bom_with_component_setuptools_with_release_notes,
    get_bom_with_component_setuptools_with_vulnerability,
    get_bom_with_component_toml_1,
    get_bom_with_dependencies_invalid,
    get_bom_with_dependencies_valid,
    get_bom_with_external_references,
    get_bom_with_metadata_component_and_dependencies,
    get_bom_with_nested_services,
    get_bom_with_services_complex,
    get_bom_with_services_simple,
)


@patch('cyclonedx.model.ThisTool._version', 'TESTING')
class TestOutputXml(BaseXmlTestCase):

    def test_bom_external_references_v1_4(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_external_references(), schema_version=SchemaVersion.V1_4,
            fixture='bom_external_references.xml'
        )

    def test_bom_external_references_v1_3(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_external_references(), schema_version=SchemaVersion.V1_3,
            fixture='bom_external_references.xml'
        )

    def test_bom_external_references_v1_2(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_external_references(), schema_version=SchemaVersion.V1_2,
            fixture='bom_external_references.xml'
        )

    def test_bom_external_references_v1_1(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_external_references(), schema_version=SchemaVersion.V1_1,
            fixture='bom_external_references.xml'
        )

    def test_bom_external_references_v1_0(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_external_references(), schema_version=SchemaVersion.V1_0,
            fixture='bom_empty.xml'
        )

    def test_simple_bom_v1_4(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_basic(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools.xml'
        )

    def test_simple_bom_v1_3(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_basic(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools.xml'
        )

    def test_simple_bom_v1_2(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_basic(), schema_version=SchemaVersion.V1_2,
            fixture='bom_setuptools.xml'
        )

    def test_simple_bom_v1_1(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_basic(), schema_version=SchemaVersion.V1_1,
            fixture='bom_setuptools.xml'
        )

    def test_simple_bom_v1_0(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_basic(), schema_version=SchemaVersion.V1_0,
            fixture='bom_setuptools.xml'
        )

    def test_simple_bom_v1_4_with_cpe(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_cpe(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools_with_cpe.xml'
        )

    def test_simple_bom_v1_3_with_cpe(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_cpe(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools_with_cpe.xml'
        )

    def test_simple_bom_v1_2_with_cpe(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_cpe(), schema_version=SchemaVersion.V1_2,
            fixture='bom_setuptools_with_cpe.xml'
        )

    def test_simple_bom_v1_1_with_cpe(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_cpe(), schema_version=SchemaVersion.V1_1,
            fixture='bom_setuptools_with_cpe.xml'
        )

    def test_simple_bom_v1_0_with_cpe(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_cpe(), schema_version=SchemaVersion.V1_0,
            fixture='bom_setuptools_with_cpe.xml'
        )

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_4)
    def test_bom_v1_4_full_component(self, mock_uuid: Mock) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_complete(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools_complete.xml'
        )
        mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_3)
    def test_bom_v1_3_full_component(self, mock_uuid: Mock) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_complete(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools_complete.xml'
        )
        mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_2)
    def test_bom_v1_2_full_component(self, mock_uuid: Mock) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_complete(), schema_version=SchemaVersion.V1_2,
            fixture='bom_setuptools_complete.xml'
        )
        mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_1)
    def test_bom_v1_1_full_component(self, mock_uuid: Mock) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_complete(), schema_version=SchemaVersion.V1_1,
            fixture='bom_setuptools_complete.xml'
        )
        mock_uuid.assert_called()

    def test_bom_v1_0_full_component(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_complete(), schema_version=SchemaVersion.V1_0,
            fixture='bom_setuptools_complete.xml'
        )

    def test_bom_v1_4_component_hashes_external_references(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_toml_1(), schema_version=SchemaVersion.V1_4,
            fixture='bom_toml_hashes_and_references.xml'
        )

    def test_bom_v1_3_component_hashes_external_references(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_toml_1(), schema_version=SchemaVersion.V1_3,
            fixture='bom_toml_hashes_and_references.xml'
        )

    def test_bom_v1_2_component_hashes_external_references(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_toml_1(), schema_version=SchemaVersion.V1_2,
            fixture='bom_toml_hashes_and_references.xml'
        )

    def test_bom_v1_1_component_hashes_external_references(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_toml_1(), schema_version=SchemaVersion.V1_1,
            fixture='bom_toml_hashes_and_references.xml'
        )

    def test_bom_v1_0_component_hashes_external_references(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_toml_1(), schema_version=SchemaVersion.V1_0,
            fixture='bom_toml_hashes_and_references.xml'
        )

    def test_bom_v1_4_no_component_version(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_no_component_version(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools_no_version.xml'
        )

    def test_bom_v1_3_no_component_version(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_no_component_version(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools_no_version.xml'
        )

    def test_bom_v1_2_no_component_version(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_no_component_version(), schema_version=SchemaVersion.V1_2,
            fixture='bom_setuptools_no_version.xml'
        )

    def test_bom_v1_1_no_component_version(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_no_component_version(), schema_version=SchemaVersion.V1_1,
            fixture='bom_setuptools_no_version.xml'
        )

    def test_bom_v1_0_no_component_version(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_no_component_version(), schema_version=SchemaVersion.V1_0,
            fixture='bom_setuptools_no_version.xml'
        )

    def test_bom_v1_4_component_with_release_notes(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_release_notes(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools_with_release_notes.xml'
        )

    def test_bom_v1_3_component_with_release_notes(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_release_notes(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools.xml'
        )

    def test_bom_v1_4_component_with_vulnerability(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_vulnerability(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools_with_vulnerabilities.xml'
        )

    def test_bom_v1_3_component_with_vulnerability(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_vulnerability(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools_with_vulnerabilities.xml'
        )

    def test_bom_v1_2_component_with_vulnerability(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_vulnerability(), schema_version=SchemaVersion.V1_2,
            fixture='bom_setuptools_with_vulnerabilities.xml'
        )

    def test_bom_v1_1_component_with_vulnerability(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_vulnerability(), schema_version=SchemaVersion.V1_1,
            fixture='bom_setuptools_with_vulnerabilities.xml'
        )

    def test_bom_v1_0_component_with_vulnerability(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_component_setuptools_with_vulnerability(), schema_version=SchemaVersion.V1_0,
            fixture='bom_setuptools.xml'
        )

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_6)
    def test_bom_v1_4_with_metadata_component(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_just_complete_metadata(), schema_version=SchemaVersion.V1_4,
                fixture='bom_with_full_metadata.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_5)
    def test_bom_v1_3_with_metadata_component(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_just_complete_metadata(), schema_version=SchemaVersion.V1_3,
                fixture='bom_with_full_metadata.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_4)
    def test_bom_v1_2_with_metadata_component(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_just_complete_metadata(), schema_version=SchemaVersion.V1_2,
                fixture='bom_with_full_metadata.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_1)
    def test_bom_v1_1_with_metadata_component(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_just_complete_metadata(), schema_version=SchemaVersion.V1_1,
                fixture='bom_empty.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_1)
    def test_bom_v1_0_with_metadata_component(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_just_complete_metadata(), schema_version=SchemaVersion.V1_0,
                fixture='bom_empty.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_4_services_simple(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_services_simple(), schema_version=SchemaVersion.V1_4,
                fixture='bom_services_simple.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_3_services_simple(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_services_simple(), schema_version=SchemaVersion.V1_3,
                fixture='bom_services_simple.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_2_services_simple(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_services_simple(), schema_version=SchemaVersion.V1_2,
                fixture='bom_services_simple.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_1_services_simple(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_services_simple(), schema_version=SchemaVersion.V1_1,
                fixture='bom_empty.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_0_services_simple(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_services_simple(), schema_version=SchemaVersion.V1_0,
                fixture='bom_empty.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_4_services_complex(self, mock_uuid4: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_services_complex(), schema_version=SchemaVersion.V1_4,
                fixture='bom_services_complex.xml'
            )
            mock_uuid4.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_3_services_complex(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_services_complex(), schema_version=SchemaVersion.V1_3,
                fixture='bom_services_complex.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_2_services_complex(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_services_complex(), schema_version=SchemaVersion.V1_2,
                fixture='bom_services_complex.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_1_services_complex(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_services_complex(), schema_version=SchemaVersion.V1_1,
                fixture='bom_empty.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_4_services_nested(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_nested_services(), schema_version=SchemaVersion.V1_4,
                fixture='bom_services_nested.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_3_services_nested(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_nested_services(), schema_version=SchemaVersion.V1_3,
                fixture='bom_services_nested.xml'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_2_services_nested(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_xml_bom(
                bom=get_bom_with_nested_services(), schema_version=SchemaVersion.V1_2,
                fixture='bom_services_nested.xml'
            )
            mock_uuid.assert_called()

    def test_bom_v1_4_dependencies(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_dependencies_valid(), schema_version=SchemaVersion.V1_4,
            fixture='bom_dependencies.xml'
        )

    def test_bom_v1_3_dependencies(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_dependencies_valid(), schema_version=SchemaVersion.V1_3,
            fixture='bom_dependencies.xml'
        )

    def test_bom_v1_2_dependencies(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_dependencies_valid(), schema_version=SchemaVersion.V1_2,
            fixture='bom_dependencies.xml'
        )

    def test_bom_v1_1_dependencies(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_dependencies_valid(), schema_version=SchemaVersion.V1_1,
            fixture='bom_dependencies.xml'
        )

    def test_bom_v1_4_dependencies_for_bom_component(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_metadata_component_and_dependencies(), schema_version=SchemaVersion.V1_4,
            fixture='bom_dependencies_component.xml'
        )

    def test_bom_v1_3_dependencies_for_bom_component(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_metadata_component_and_dependencies(), schema_version=SchemaVersion.V1_3,
            fixture='bom_dependencies_component.xml'
        )

    def test_bom_v1_2_dependencies_for_bom_component(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_with_metadata_component_and_dependencies(), schema_version=SchemaVersion.V1_2,
            fixture='bom_dependencies_component.xml'
        )

    def test_bom_v1_4_dependencies_invalid(self) -> None:
        with self.assertRaises(UnknownComponentDependencyException):
            self._validate_xml_bom(
                bom=get_bom_with_dependencies_invalid(), schema_version=SchemaVersion.V1_4,
                fixture='bom_dependencies.xml'
            )

    def test_bom_v1_4_issue_275_components(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_for_issue_275_components(), schema_version=SchemaVersion.V1_4,
            fixture='bom_issue_275_components.xml'
        )

    def test_bom_v1_3_issue_275_components(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_for_issue_275_components(), schema_version=SchemaVersion.V1_3,
            fixture='bom_issue_275_components.xml'
        )

    def test_bom_v1_2_issue_275_components(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_for_issue_275_components(), schema_version=SchemaVersion.V1_2,
            fixture='bom_issue_275_components.xml'
        )

    def test_bom_v1_1_issue_275_components(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_for_issue_275_components(), schema_version=SchemaVersion.V1_1,
            fixture='bom_issue_275_components.xml'
        )

    def test_bom_v1_0_issue_275_components(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_for_issue_275_components(), schema_version=SchemaVersion.V1_0,
            fixture='bom_issue_275_components.xml'
        )

    def test_bom_v1_4_issue_328_components(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_for_issue_328_components(), schema_version=SchemaVersion.V1_4,
            fixture='bom_issue_328_components.xml'
        )

    def test_bom_v1_3_issue_328_components(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_for_issue_328_components(), schema_version=SchemaVersion.V1_3,
            fixture='bom_issue_328_components.xml'
        )

    def test_bom_v1_2_issue_328_components(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_for_issue_328_components(), schema_version=SchemaVersion.V1_2,
            fixture='bom_issue_328_components.xml'
        )

    def test_bom_v1_1_issue_328_components(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_for_issue_328_components(), schema_version=SchemaVersion.V1_1,
            fixture='bom_issue_328_components.xml'
        )

    def test_bom_v1_0_issue_328_components(self) -> None:
        self._validate_xml_bom(
            bom=get_bom_for_issue_328_components(), schema_version=SchemaVersion.V1_0,
            fixture='bom_issue_328_components.xml'
        )

    # region Helper methods

    def _validate_xml_bom(self, bom: Bom, schema_version: SchemaVersion, fixture: str) -> None:
        outputter = get_instance(bom=bom, schema_version=schema_version)
        self.assertEqual(outputter.schema_version, schema_version)
        output = outputter.output_as_string()
        self.assertValidAgainstSchema(bom_xml=output, schema_version=schema_version)
        fixture_file = join(
            FIXTURES_DIRECTORY,
            'xml',
            schema_version.to_version(),
            fixture
        )
        if RECREATE_SNAPSHOTS:
            with open(fixture_file, 'w') as expected_xml:
                expected_xml.write(output)
        with open(fixture_file) as expected_xml:
            self.assertEqualXmlBom(expected_xml.read(), output, namespace=outputter.get_target_namespace())

    # endregion Helper methods



# encoding: utf-8

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

"""
Set of classes and methods for outputting our libraries internal Bom model to CycloneDX documents in varying formats
and according to different versions of the CycloneDX schema standard.
"""

import importlib
import os
from abc import ABC, abstractmethod
from enum import Enum
from typing import Iterable, Union, cast

from ..model.bom import Bom
from ..model.component import Component


class OutputFormat(str, Enum):
    JSON: str = 'Json'
    XML: str = 'Xml'


class SchemaVersion(str, Enum):
    V1_0: str = 'V1Dot0'
    V1_1: str = 'V1Dot1'
    V1_2: str = 'V1Dot2'
    V1_3: str = 'V1Dot3'
    V1_4: str = 'V1Dot4'

    def to_version(self) -> str:
        """
        Return as a version string - e.g. `1.4`

        Returns:
            `str` version
        """
        return f'{self.value[1]}.{self.value[5]}'


LATEST_SUPPORTED_SCHEMA_VERSION = SchemaVersion.V1_4


class BaseOutput(ABC):

    def __init__(self, bom: Bom, **kwargs: int) -> None:
        super().__init__(**kwargs)
        self._bom = bom
        self._generated: bool = False

    def _chained_components(self, container: Union[Bom, Component]) -> Iterable[Component]:
        for component in container.components:
            yield component
            yield from self._chained_components(component)

    @property
    @abstractmethod
    def schema_version(self) -> SchemaVersion:
        pass

    @property
    def generated(self) -> bool:
        return self._generated

    @generated.setter
    def generated(self, generated: bool) -> None:
        self._generated = generated

    def get_bom(self) -> Bom:
        return self._bom

    def set_bom(self, bom: Bom) -> None:
        self._bom = bom

    @abstractmethod
    def generate(self, force_regeneration: bool = False) -> None:
        pass

    @abstractmethod
    def output_as_string(self) -> str:
        pass

    def output_to_file(self, filename: str, allow_overwrite: bool = False) -> None:
        # Check directory writable
        output_filename = os.path.realpath(filename)
        output_directory = os.path.dirname(output_filename)

        if not os.access(output_directory, os.W_OK):
            raise PermissionError(output_directory)

        if os.path.exists(output_filename) and not allow_overwrite:
            raise FileExistsError(output_filename)

        with open(output_filename, mode='wb') as f_out:
            f_out.write(self.output_as_string().encode('utf-8'))

        f_out.close()


def get_instance(bom: Bom, output_format: OutputFormat = OutputFormat.XML,
                 schema_version: SchemaVersion = LATEST_SUPPORTED_SCHEMA_VERSION) -> BaseOutput:
    """
    Helper method to quickly get the correct output class/formatter.

    Pass in your BOM and optionally an output format and schema version (defaults to XML and latest schema version).

    :param bom: Bom
    :param output_format: OutputFormat
    :param schema_version: SchemaVersion
    :return:
    """
    try:
        module = importlib.import_module(f"cyclonedx.output.{output_format.value.lower()}")
        output_klass = getattr(module, f"{output_format.value}{schema_version.value}")
    except (ImportError, AttributeError):
        raise ValueError(f"Unknown format {output_format.value.lower()!r}") from None

    return cast(BaseOutput, output_klass(bom=bom))



# encoding: utf-8

# This file is part of CycloneDX Python Lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

from os.path import join
from unittest.mock import Mock, patch

from cyclonedx.exception.model import UnknownComponentDependencyException
from cyclonedx.exception.output import FormatNotSupportedException
from cyclonedx.model.bom import Bom
from cyclonedx.output import OutputFormat, SchemaVersion, get_instance

from . import FIXTURES_DIRECTORY, RECREATE_SNAPSHOTS
from .base import BaseJsonTestCase
from .data import (
    MOCK_UUID_1,
    MOCK_UUID_2,
    MOCK_UUID_3,
    MOCK_UUID_6,
    TEST_UUIDS,
    get_bom_for_issue_275_components,
    get_bom_for_issue_328_components,
    get_bom_just_complete_metadata,
    get_bom_with_component_setuptools_basic,
    get_bom_with_component_setuptools_complete,
    get_bom_with_component_setuptools_no_component_version,
    get_bom_with_component_setuptools_with_cpe,
    get_bom_with_component_setuptools_with_release_notes,
    get_bom_with_component_setuptools_with_vulnerability,
    get_bom_with_component_toml_1,
    get_bom_with_dependencies_invalid,
    get_bom_with_dependencies_valid,
    get_bom_with_external_references,
    get_bom_with_metadata_component_and_dependencies,
    get_bom_with_nested_services,
    get_bom_with_services_complex,
    get_bom_with_services_simple,
)


@patch('cyclonedx.model.ThisTool._version', 'TESTING')
class TestOutputJson(BaseJsonTestCase):

    def test_bom_external_references_v1_4(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_external_references(), schema_version=SchemaVersion.V1_4,
            fixture='bom_external_references.json'
        )

    def test_bom_external_references_v1_3(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_external_references(), schema_version=SchemaVersion.V1_3,
            fixture='bom_external_references.json'
        )

    def test_bom_external_references_v1_2(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_external_references(), schema_version=SchemaVersion.V1_2,
            fixture='bom_external_references.json'
        )

    def test_simple_bom_v1_4(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_basic(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools.json'
        )

    def test_simple_bom_v1_3(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_basic(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools.json'
        )

    def test_simple_bom_v1_2(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_basic(), schema_version=SchemaVersion.V1_2,
            fixture='bom_setuptools.json'
        )

    def test_simple_bom_v1_1(self) -> None:
        self._validate_json_bom_not_supported(
            bom=get_bom_with_component_setuptools_basic(), schema_version=SchemaVersion.V1_1
        )

    def test_simple_bom_v1_0(self) -> None:
        self._validate_json_bom_not_supported(
            bom=get_bom_with_component_setuptools_basic(), schema_version=SchemaVersion.V1_0
        )

    def test_simple_bom_v1_4_with_cpe(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_with_cpe(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools_with_cpe.json'
        )

    def test_simple_bom_v1_3_with_cpe(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_with_cpe(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools_with_cpe.json'
        )

    def test_simple_bom_v1_2_with_cpe(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_with_cpe(), schema_version=SchemaVersion.V1_2,
            fixture='bom_setuptools_with_cpe.json'
        )

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_6)
    def test_bom_v1_4_full_component(self, mock: Mock) -> None:
        self.maxDiff = None
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_complete(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools_complete.json'
        )
        mock.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_6)
    def test_bom_v1_3_full_component(self, mock: Mock) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_complete(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools_complete.json'
        )
        mock.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_6)
    def test_bom_v1_2_full_component(self, mock: Mock) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_complete(), schema_version=SchemaVersion.V1_2,
            fixture='bom_setuptools_complete.json'
        )
        mock.assert_called()

    def test_bom_v1_4_component_hashes_external_references(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_toml_1(), schema_version=SchemaVersion.V1_4,
            fixture='bom_toml_1.json'
        )

    def test_bom_v1_3_component_hashes_external_references(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_toml_1(), schema_version=SchemaVersion.V1_3,
            fixture='bom_toml_1.json'
        )

    def test_bom_v1_2_component_hashes_external_references(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_toml_1(), schema_version=SchemaVersion.V1_2,
            fixture='bom_toml_1.json'
        )

    def test_bom_v1_1_component_hashes_external_references(self) -> None:
        self._validate_json_bom_not_supported(bom=get_bom_with_component_toml_1(), schema_version=SchemaVersion.V1_1)

    def test_bom_v1_0_component_hashes_external_references(self) -> None:
        self._validate_json_bom_not_supported(bom=get_bom_with_component_toml_1(), schema_version=SchemaVersion.V1_0)

    def test_bom_v1_4_no_component_version(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_no_component_version(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools_no_version.json'
        )

    def test_bom_v1_3_no_component_version(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_no_component_version(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools_no_version.json'
        )

    def test_bom_v1_4_component_with_release_notes(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_with_release_notes(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools_with_release_notes.json'
        )

    def test_bom_v1_3_component_with_release_notes(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_with_release_notes(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools.json'
        )

    def test_bom_v1_4_component_with_vulnerability(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_with_vulnerability(), schema_version=SchemaVersion.V1_4,
            fixture='bom_setuptools_with_vulnerabilities.json'
        )

    def test_bom_v1_3_component_with_vulnerability(self) -> None:
        # Vulnerabilities not support in JSON < 1.4
        self._validate_json_bom(
            bom=get_bom_with_component_setuptools_with_vulnerability(), schema_version=SchemaVersion.V1_3,
            fixture='bom_setuptools.json'
        )

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_1)
    def test_bom_v1_4_with_metadata_component(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_just_complete_metadata(), schema_version=SchemaVersion.V1_4,
                fixture='bom_with_full_metadata.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_2)
    def test_bom_v1_3_with_metadata_component(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_just_complete_metadata(), schema_version=SchemaVersion.V1_3,
                fixture='bom_with_full_metadata.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', return_value=MOCK_UUID_3)
    def test_bom_v1_2_with_metadata_component(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_just_complete_metadata(), schema_version=SchemaVersion.V1_2,
                fixture='bom_with_full_metadata.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_4_services_simple(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_with_services_simple(), schema_version=SchemaVersion.V1_4,
                fixture='bom_services_simple.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_3_services_simple(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_with_services_simple(), schema_version=SchemaVersion.V1_3,
                fixture='bom_services_simple.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_2_services_simple(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_with_services_simple(), schema_version=SchemaVersion.V1_2,
                fixture='bom_services_simple.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_4_services_complex(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_with_services_complex(), schema_version=SchemaVersion.V1_4,
                fixture='bom_services_complex.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_3_services_complex(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_with_services_complex(), schema_version=SchemaVersion.V1_3,
                fixture='bom_services_complex.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_2_services_complex(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_with_services_complex(), schema_version=SchemaVersion.V1_2,
                fixture='bom_services_complex.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_4_services_nested(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_with_nested_services(), schema_version=SchemaVersion.V1_4,
                fixture='bom_services_nested.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_3_services_nested(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_with_nested_services(), schema_version=SchemaVersion.V1_3,
                fixture='bom_services_nested.json'
            )
            mock_uuid.assert_called()

    @patch('cyclonedx.model.bom_ref.uuid4', side_effect=TEST_UUIDS)
    def test_bom_v1_2_services_nested(self, mock_uuid: Mock) -> None:
        with self.assertWarns(UserWarning):
            self._validate_json_bom(
                bom=get_bom_with_nested_services(), schema_version=SchemaVersion.V1_2,
                fixture='bom_services_nested.json'
            )
            mock_uuid.assert_called()

    def test_bom_v1_4_dependencies(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_dependencies_valid(), schema_version=SchemaVersion.V1_4,
            fixture='bom_dependencies.json'
        )

    def test_bom_v1_3_dependencies(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_dependencies_valid(), schema_version=SchemaVersion.V1_3,
            fixture='bom_dependencies.json'
        )

    def test_bom_v1_2_dependencies(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_dependencies_valid(), schema_version=SchemaVersion.V1_2,
            fixture='bom_dependencies.json'
        )

    def test_bom_v1_4_dependencies_for_bom_component(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_metadata_component_and_dependencies(), schema_version=SchemaVersion.V1_4,
            fixture='bom_dependencies_component.json'
        )

    def test_bom_v1_3_dependencies_for_bom_component(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_metadata_component_and_dependencies(), schema_version=SchemaVersion.V1_3,
            fixture='bom_dependencies_component.json'
        )

    def test_bom_v1_2_dependencies_for_bom_component(self) -> None:
        self._validate_json_bom(
            bom=get_bom_with_metadata_component_and_dependencies(), schema_version=SchemaVersion.V1_2,
            fixture='bom_dependencies_component.json'
        )

    def test_bom_v1_4_dependencies_invalid(self) -> None:
        with self.assertRaises(UnknownComponentDependencyException):
            self._validate_json_bom(
                bom=get_bom_with_dependencies_invalid(), schema_version=SchemaVersion.V1_4,
                fixture='bom_dependencies.json'
            )

    def test_bom_v1_4_issue_275_components(self) -> None:
        self._validate_json_bom(
            bom=get_bom_for_issue_275_components(), schema_version=SchemaVersion.V1_4,
            fixture='bom_issue_275_components.json'
        )

    def test_bom_v1_3_issue_275_components(self) -> None:
        self._validate_json_bom(
            bom=get_bom_for_issue_275_components(), schema_version=SchemaVersion.V1_3,
            fixture='bom_issue_275_components.json'
        )

    def test_bom_v1_2_issue_275_components(self) -> None:
        self._validate_json_bom(
            bom=get_bom_for_issue_275_components(), schema_version=SchemaVersion.V1_2,
            fixture='bom_issue_275_components.json'
        )

    def test_bom_v1_4_issue_328_components(self) -> None:
        self._validate_json_bom(
            bom=get_bom_for_issue_328_components(), schema_version=SchemaVersion.V1_4,
            fixture='bom_issue_328_components.json'
        )

    def test_bom_v1_3_issue_328_components(self) -> None:
        self._validate_json_bom(
            bom=get_bom_for_issue_328_components(), schema_version=SchemaVersion.V1_3,
            fixture='bom_issue_328_components.json'
        )

    def test_bom_v1_2_issue_328_components(self) -> None:
        self._validate_json_bom(
            bom=get_bom_for_issue_328_components(), schema_version=SchemaVersion.V1_2,
            fixture='bom_issue_328_components.json'
        )

    # region Helper methods

    def _validate_json_bom(self, bom: Bom, schema_version: SchemaVersion, fixture: str) -> None:
        outputter = get_instance(bom=bom, output_format=OutputFormat.JSON, schema_version=schema_version)
        self.assertEqual(outputter.schema_version, schema_version)
        output = outputter.output_as_string()
        self.assertValidAgainstSchema(bom_json=output, schema_version=schema_version)
        fixture_file = join(
            FIXTURES_DIRECTORY,
            'json',
            schema_version.to_version(),
            fixture
        )
        if RECREATE_SNAPSHOTS:
            with open(fixture_file, 'w') as expected_json:
                expected_json.write(output)
        with open(fixture_file) as expected_json:
            self.assertEqualJsonBom(expected_json.read(), output)

    def _validate_json_bom_not_supported(self, bom: Bom, schema_version: SchemaVersion) -> None:
        with self.assertRaises(FormatNotSupportedException):
            outputter = get_instance(bom=bom, output_format=OutputFormat.JSON, schema_version=schema_version)
            outputter.output_as_string()

    # endregion Helper methods



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cyclonedx',
 'cyclonedx.exception',
 'cyclonedx.factory',
 'cyclonedx.model',
 'cyclonedx.output',
 'cyclonedx.output.serializer',
 'cyclonedx.parser']

package_data = \
{'': ['*'], 'cyclonedx': ['schema/*', 'schema/ext/*']}

install_requires = \
['packageurl-python>=0.9',
 'setuptools>=47.0.0',
 'sortedcontainers>=2.4.0,<3.0.0',
 'toml>=0.10.0,<0.11.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=3.4']}

setup_kwargs = {
    'name': 'cyclonedx-python-lib',
    'version': '3.1.4',
    'description': 'A library for producing CycloneDX SBOM (Software Bill of Materials) files.',
    'long_description': '# Python Library for generating CycloneDX\n\n[![shield_gh-workflow-test]][link_gh-workflow-test]\n[![shield_rtfd]][link_rtfd]\n[![shield_pypi-version]][link_pypi]\n[![shield_conda-forge-version]][link_conda-forge]\n[![shield_license]][license_file]  \n[![shield_website]][link_website]\n[![shield_slack]][link_slack]\n[![shield_groups]][link_discussion]\n[![shield_twitter-follow]][link_twitter]\n\n----\n\nThis CycloneDX module for Python can generate valid CycloneDX bill-of-material document containing an aggregate of all\nproject dependencies.\n\nThis module is not designed for standalone use.\n\nIf you\'re looking for a CycloneDX tool to run to generate (SBOM) software bill-of-materials documents, why not checkout: [CycloneDX Python][cyclonedx-python]\n\nAdditionally, the following tool can be used as well (and this library was written to help improve it) [Jake][jake].\n\nAdditionally, you can use this module yourself in your application to programmatically generate SBOMs.\n\nCycloneDX is a lightweight BOM specification that is easily created, human-readable, and simple to parse.\n\nView our documentation [here](https://cyclonedx-python-library.readthedocs.io/).\n\n## Python Support\n\nWe endeavour to support all functionality for all [current actively supported Python versions](https://www.python.org/downloads/).\nHowever, some features may not be possible/present in older Python versions due to their lack of support.\n\n## Changelog\n\nSee our [CHANGELOG][chaneglog_file].\n\n## Contributing\n\nFeel free to open issues, bugreports or pull requests.  \nSee the [CONTRIBUTING][contributing_file] file for details.\n\n## Copyright & License\n\nCycloneDX Python Lib is Copyright (c) OWASP Foundation. All Rights Reserved.  \nPermission to modify and redistribute is granted under the terms of the Apache 2.0 license.  \nSee the [LICENSE][license_file] file for the full license.\n\n[cyclonedx-python]: https://github.com/CycloneDX/cyclonedx-python\n[jake]: https://github.com/sonatype-nexus-community/jake\n\n[license_file]: https://github.com/CycloneDX/cyclonedx-python-lib/blob/master/LICENSE\n[chaneglog_file]: https://github.com/CycloneDX/cyclonedx-python-lib/blob/master/CHANGELOG.md\n[contributing_file]: https://github.com/CycloneDX/cyclonedx-python-lib/blob/master/CONTRIBUTING.md\n\n[shield_gh-workflow-test]: https://img.shields.io/github/actions/workflow/status/CycloneDX/cyclonedx-python-lib/poetry.yml?branch=main&logo=GitHub&logoColor=white "build"\n[shield_pypi-version]: https://img.shields.io/pypi/v/cyclonedx-python-lib?logo=pypi&logoColor=white&label=PyPI "PyPI"\n[shield_conda-forge-version]: https://img.shields.io/conda/vn/conda-forge/cyclonedx-python-lib?logo=anaconda&logoColor=white&label=conda-forge "conda-forge"\n[shield_rtfd]: https://img.shields.io/readthedocs/cyclonedx-python-library?logo=readthedocs&logoColor=white "Read the Docs"\n[shield_license]: https://img.shields.io/github/license/CycloneDX/cyclonedx-python-lib?logo=open%20source%20initiative&logoColor=white "license"\n[shield_website]: https://img.shields.io/badge/https://-cyclonedx.org-blue.svg "homepage"\n[shield_slack]: https://img.shields.io/badge/slack-join-blue?logo=Slack&logoColor=white "slack join"\n[shield_groups]: https://img.shields.io/badge/discussion-groups.io-blue.svg "groups discussion"\n[shield_twitter-follow]: https://img.shields.io/badge/Twitter-follow-blue?logo=Twitter&logoColor=white "twitter follow"\n[link_gh-workflow-test]: https://github.com/CycloneDX/cyclonedx-python-lib/actions/workflows/poetry.yml?query=branch%3Amain\n[link_pypi]: https://pypi.org/project/cyclonedx-python-lib/\n[link_conda-forge]: https://anaconda.org/conda-forge/cyclonedx-python-lib\n[link_rtfd]: https://cyclonedx-python-library.readthedocs.io/en/latest/?badge=latest\n[link_website]: https://cyclonedx.org/\n[link_slack]: https://cyclonedx.org/slack/invite\n[link_discussion]: https://groups.io/g/CycloneDX\n[link_twitter]: https://twitter.com/CycloneDX_Spec\n\n[PEP-508]: https://www.python.org/dev/peps/pep-0508/\n',
    'author': 'Paul Horton',
    'author_email': 'phorton@sonatype.com',
    'maintainer': 'Paul Horton',
    'maintainer_email': 'phorton@sonatype.com',
    'url': 'https://github.com/CycloneDX/cyclonedx-python-lib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# encoding: utf-8

# This file is part of CycloneDX Python Lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

from unittest import TestCase

from cyclonedx.exception.model import NoPropertiesProvidedException
from cyclonedx.model import XsUri
from cyclonedx.model.issue import IssueClassification, IssueType, IssueTypeSource

from .data import get_issue_1, get_issue_2, reorder


class TestModelIssueType(TestCase):

    def test_same(self) -> None:
        i_1 = get_issue_1()
        i_2 = get_issue_1()
        self.assertNotEqual(id(i_1), id(i_2))
        self.assertEqual(hash(i_1), hash(i_2))
        self.assertTrue(i_1 == i_2)

    def test_not_same(self) -> None:
        i_1 = get_issue_1()
        i_2 = get_issue_2()
        self.assertNotEqual(id(i_1), id(i_2))
        self.assertNotEqual(hash(i_1), hash(i_2))
        self.assertFalse(i_1 == i_2)

    def test_sort(self) -> None:
        source_a = IssueTypeSource(name='a')
        source_b = IssueTypeSource(name='b')

        # expected sort order: (type/classification, id, name, description, source)
        expected_order = [6, 5, 0, 1, 2, 3, 4]
        issues = [
            IssueType(classification=IssueClassification.SECURITY, id_='a', name='a', description='a', source=source_a),
            IssueType(classification=IssueClassification.SECURITY, id_='a', name='a', description='a', source=source_b),
            IssueType(classification=IssueClassification.SECURITY, id_='a', name='a', description='a'),
            IssueType(classification=IssueClassification.SECURITY, id_='a', name='a'),
            IssueType(classification=IssueClassification.SECURITY, id_='a'),
            IssueType(classification=IssueClassification.DEFECT, id_='a', name='a', description='a', source=source_b),
            IssueType(classification=IssueClassification.DEFECT, id_='a', name='a', description='a', source=source_a),
        ]
        sorted_issues = sorted(issues)
        expected_issues = reorder(issues, expected_order)
        self.assertListEqual(sorted_issues, expected_issues)


class TestModelIssueTypeSource(TestCase):

    def test_no_params(self) -> None:
        with self.assertRaises(NoPropertiesProvidedException):
            IssueTypeSource()

    def test_same(self) -> None:
        its_1 = IssueTypeSource(name="The Source", url=XsUri('https://cyclonedx.org'))
        its_2 = IssueTypeSource(name="The Source", url=XsUri('https://cyclonedx.org'))
        self.assertNotEqual(id(its_1), id(its_2))
        self.assertEqual(hash(its_1), hash(its_2))
        self.assertTrue(its_1 == its_2)

    def test_not_same(self) -> None:
        its_1 = IssueTypeSource(name="The Source", url=XsUri('https://cyclonedx.org'))
        its_2 = IssueTypeSource(name="Not the Source", url=XsUri('https://cyclonedx.org'))
        self.assertNotEqual(id(its_1), id(its_2))
        self.assertNotEqual(hash(its_1), hash(its_2))
        self.assertFalse(its_1 == its_2)

    def test_sort(self) -> None:
        # expected sort order: ([name], [url])
        expected_order = [0, 1, 3, 2, 5, 4]
        sources = [
            IssueTypeSource(name='a', url=XsUri('a')),
            IssueTypeSource(name='a', url=XsUri('b')),
            IssueTypeSource(name='b'),
            IssueTypeSource(name='a'),
            IssueTypeSource(url=XsUri('b')),
            IssueTypeSource(url=XsUri('a')),
        ]
        sorted_sources = sorted(sources)
        expected_sources = reorder(sources, expected_order)
        self.assertListEqual(sorted_sources, expected_sources)



# encoding: utf-8

# This file is part of CycloneDX Python Lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

from unittest import TestCase
from unittest.mock import Mock, patch

from cyclonedx.model.service import Service

from .data import reorder


class TestModelService(TestCase):

    @patch('cyclonedx.model.bom_ref.uuid4', return_value='77d15ab9-5602-4cca-8ed2-59ae579aafd3')
    def test_minimal_service(self, mock_uuid: Mock) -> None:
        s = Service(name='my-test-service')
        mock_uuid.assert_called()
        self.assertEqual(s.name, 'my-test-service')
        self.assertEqual(str(s.bom_ref), '77d15ab9-5602-4cca-8ed2-59ae579aafd3')
        self.assertIsNone(s.provider)
        self.assertIsNone(s.group)
        self.assertIsNone(s.version)
        self.assertIsNone(s.description)
        self.assertFalse(s.endpoints)
        self.assertIsNone(s.authenticated)
        self.assertIsNone(s.x_trust_boundary)
        self.assertFalse(s.data)
        self.assertFalse(s.licenses)
        self.assertFalse(s.external_references)
        self.assertFalse(s.services)
        self.assertFalse(s.release_notes)
        self.assertFalse(s.properties)

    @patch('cyclonedx.model.bom_ref.uuid4', return_value='859ff614-35a7-4d37-803b-d89130cb2577')
    def test_service_with_services(self, mock_uuid: Mock) -> None:
        parent_service = Service(name='parent-service')
        parent_service.services = [
            Service(name='child-service-1'),
            Service(name='child-service-2')
        ]
        mock_uuid.assert_called()
        self.assertEqual(parent_service.name, 'parent-service')
        self.assertEqual(str(parent_service.bom_ref), '859ff614-35a7-4d37-803b-d89130cb2577')
        self.assertIsNone(parent_service.provider)
        self.assertIsNone(parent_service.group)
        self.assertIsNone(parent_service.version)
        self.assertIsNone(parent_service.description)
        self.assertFalse(parent_service.endpoints)
        self.assertIsNone(parent_service.authenticated)
        self.assertIsNone(parent_service.x_trust_boundary)
        self.assertFalse(parent_service.data)
        self.assertFalse(parent_service.licenses)
        self.assertFalse(parent_service.external_references)
        self.assertIsNotNone(parent_service.services)
        self.assertEqual(len(parent_service.services), 2)
        self.assertIsNone(parent_service.release_notes)
        self.assertFalse(parent_service.properties)
        self.assertTrue(Service(name='child-service-1') in parent_service.services)

    def test_sort(self) -> None:
        # expected sort order: ([group], name, [version])
        expected_order = [0, 1, 3, 4, 2, 5]
        services = [
            Service(name='service-a', group='group-a'),
            Service(name='service-b', group='group-a', version='1.0.0'),
            Service(name='service-c', version='2.0.0'),
            Service(name='service-b', group='group-a'),
            Service(name='service-c', version='1.0.0'),
            Service(name='service-d', ),
        ]
        sorted_services = sorted(services)
        expected_services = reorder(services, expected_order)
        self.assertListEqual(sorted_services, expected_services)



# encoding: utf-8

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

from typing import Any, Iterable, Optional

from sortedcontainers import SortedSet

from . import (
    ComparableTuple,
    DataClassification,
    ExternalReference,
    LicenseChoice,
    OrganizationalEntity,
    Property,
    XsUri,
)
from .bom_ref import BomRef
from .release_note import ReleaseNotes

"""
This set of classes represents the data that is possible about known Services.

.. note::
    See the CycloneDX Schema extension definition https://cyclonedx.org/docs/1.4/xml/#type_servicesType
"""


class Service:
    """
    Class that models the `service` complex type in the CycloneDX schema.

    .. note::
        See the CycloneDX schema: https://cyclonedx.org/docs/1.4/xml/#type_service
    """

    def __init__(self, *, name: str, bom_ref: Optional[str] = None, provider: Optional[OrganizationalEntity] = None,
                 group: Optional[str] = None, version: Optional[str] = None, description: Optional[str] = None,
                 endpoints: Optional[Iterable[XsUri]] = None, authenticated: Optional[bool] = None,
                 x_trust_boundary: Optional[bool] = None, data: Optional[Iterable[DataClassification]] = None,
                 licenses: Optional[Iterable[LicenseChoice]] = None,
                 external_references: Optional[Iterable[ExternalReference]] = None,
                 properties: Optional[Iterable[Property]] = None,
                 services: Optional[Iterable['Service']] = None,
                 release_notes: Optional[ReleaseNotes] = None,
                 ) -> None:
        self._bom_ref = BomRef(value=bom_ref)
        self.provider = provider
        self.group = group
        self.name = name
        self.version = version
        self.description = description
        self.endpoints = endpoints or []  # type: ignore
        self.authenticated = authenticated
        self.x_trust_boundary = x_trust_boundary
        self.data = data or []  # type: ignore
        self.licenses = licenses or []  # type: ignore
        self.external_references = external_references or []  # type: ignore
        self.services = services or []  # type: ignore
        self.release_notes = release_notes
        self.properties = properties or []  # type: ignore

    @property
    def bom_ref(self) -> BomRef:
        """
        An optional identifier which can be used to reference the service elsewhere in the BOM. Uniqueness is enforced
        within all elements and children of the root-level bom element.

        If a value was not provided in the constructor, a UUIDv4 will have been assigned.

        Returns:
           `BomRef` unique identifier for this Service
        """
        return self._bom_ref

    @property
    def provider(self) -> Optional[OrganizationalEntity]:
        """
        Get the The organization that provides the service.

        Returns:
            `OrganizationalEntity` if set else `None`
        """
        return self._provider

    @provider.setter
    def provider(self, provider: Optional[OrganizationalEntity]) -> None:
        self._provider = provider

    @property
    def group(self) -> Optional[str]:
        """
        The grouping name, namespace, or identifier. This will often be a shortened, single name of the company or
        project that produced the service or domain name. Whitespace and special characters should be avoided.

        Returns:
            `str` if provided else `None`
        """
        return self._group

    @group.setter
    def group(self, group: Optional[str]) -> None:
        self._group = group

    @property
    def name(self) -> str:
        """
        The name of the service. This will often be a shortened, single name of the service.

        Returns:
            `str`
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def version(self) -> Optional[str]:
        """
        The service version.

        Returns:
            `str` if set else `None`
        """
        return self._version

    @version.setter
    def version(self, version: Optional[str]) -> None:
        self._version = version

    @property
    def description(self) -> Optional[str]:
        """
        Specifies a description for the service.

        Returns:
            `str` if set else `None`
        """
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self._description = description

    @property
    def endpoints(self) -> "SortedSet[XsUri]":
        """
        A list of endpoints URI's this service provides.

        Returns:
            Set of `XsUri`
        """
        return self._endpoints

    @endpoints.setter
    def endpoints(self, endpoints: Iterable[XsUri]) -> None:
        self._endpoints = SortedSet(endpoints)

    @property
    def authenticated(self) -> Optional[bool]:
        """
        A boolean value indicating if the service requires authentication. A value of true indicates the service
        requires authentication prior to use.

        A value of false indicates the service does not require authentication.

        Returns:
            `bool` if set else `None`
        """
        return self._authenticated

    @authenticated.setter
    def authenticated(self, authenticated: Optional[bool]) -> None:
        self._authenticated = authenticated

    @property
    def x_trust_boundary(self) -> Optional[bool]:
        """
        A boolean value indicating if use of the service crosses a trust zone or boundary. A value of true indicates
        that by using the service, a trust boundary is crossed.

        A value of false indicates that by using the service, a trust boundary is not crossed.

        Returns:
            `bool` if set else `None`
        """
        return self._x_trust_boundary

    @x_trust_boundary.setter
    def x_trust_boundary(self, x_trust_boundary: Optional[bool]) -> None:
        self._x_trust_boundary = x_trust_boundary

    @property
    def data(self) -> "SortedSet[DataClassification]":
        """
        Specifies the data classification.

        Returns:
            Set of `DataClassification`
        """
        return self._data

    @data.setter
    def data(self, data: Iterable[DataClassification]) -> None:
        self._data = SortedSet(data)

    @property
    def licenses(self) -> "SortedSet[LicenseChoice]":
        """
        A optional list of statements about how this Service is licensed.

        Returns:
            Set of `LicenseChoice`
        """
        return self._licenses

    @licenses.setter
    def licenses(self, licenses: Iterable[LicenseChoice]) -> None:
        self._licenses = SortedSet(licenses)

    @property
    def external_references(self) -> "SortedSet[ExternalReference]":
        """
        Provides the ability to document external references related to the Service.

        Returns:
            Set of `ExternalReference`
        """
        return self._external_references

    @external_references.setter
    def external_references(self, external_references: Iterable[ExternalReference]) -> None:
        self._external_references = SortedSet(external_references)

    @property
    def services(self) -> "SortedSet['Service']":
        """
        A list of services included or deployed behind the parent service.

        This is not a dependency tree.

        It provides a way to specify a hierarchical representation of service assemblies.

        Returns:
            Set of `Service`
        """
        return self._services

    @services.setter
    def services(self, services: Iterable['Service']) -> None:
        self._services = SortedSet(services)

    @property
    def release_notes(self) -> Optional[ReleaseNotes]:
        """
        Specifies optional release notes.

        Returns:
            `ReleaseNotes` or `None`
        """
        return self._release_notes

    @release_notes.setter
    def release_notes(self, release_notes: Optional[ReleaseNotes]) -> None:
        self._release_notes = release_notes

    @property
    def properties(self) -> "SortedSet[Property]":
        """
        Provides the ability to document properties in a key/value store. This provides flexibility to include data not
        officially supported in the standard without having to use additional namespaces or create extensions.

        Return:
            Set of `Property`
        """
        return self._properties

    @properties.setter
    def properties(self, properties: Iterable[Property]) -> None:
        self._properties = SortedSet(properties)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Service):
            return hash(other) == hash(self)
        return False

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, Service):
            return ComparableTuple((self.group, self.name, self.version)) < \
                ComparableTuple((other.group, other.name, other.version))
        return NotImplemented

    def __hash__(self) -> int:
        return hash((
            self.authenticated, tuple(self.data), self.description, tuple(self.endpoints),
            tuple(self.external_references), self.group, tuple(self.licenses), self.name, tuple(self.properties),
            self.provider, self.release_notes, tuple(self.services), self.version, self.x_trust_boundary
        ))

    def __repr__(self) -> str:
        return f'<Service group={self.group}, name={self.name}, version={self.version}, bom-ref={self.bom_ref}>'



# encoding: utf-8

# This file is part of CycloneDX Python Lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

from datetime import datetime
from typing import Iterable, Optional

from sortedcontainers import SortedSet

from ..model import Note, Property, XsUri
from ..model.issue import IssueType


class ReleaseNotes:
    """
    This is our internal representation of a `releaseNotesType` for a Component in a BOM.

    .. note::
        See the CycloneDX Schema definition: https://cyclonedx.org/docs/1.4/#type_releaseNotesType
    """

    def __init__(self, *, type_: str, title: Optional[str] = None, featured_image: Optional[XsUri] = None,
                 social_image: Optional[XsUri] = None, description: Optional[str] = None,
                 timestamp: Optional[datetime] = None, aliases: Optional[Iterable[str]] = None,
                 tags: Optional[Iterable[str]] = None, resolves: Optional[Iterable[IssueType]] = None,
                 notes: Optional[Iterable[Note]] = None, properties: Optional[Iterable[Property]] = None) -> None:
        self.type = type_
        self.title = title
        self.featured_image = featured_image
        self.social_image = social_image
        self.description = description
        self.timestamp = timestamp
        self.aliases = aliases or []  # type: ignore
        self.tags = tags or []  # type: ignore
        self.resolves = resolves or []  # type: ignore
        self.notes = notes or []  # type: ignore
        self.properties = properties or []  # type: ignore

    @property
    def type(self) -> str:
        """
        The software versioning type.

        It is **RECOMMENDED** that the release type use one of 'major', 'minor', 'patch', 'pre-release', or 'internal'.

        Representing all possible software release types is not practical, so standardizing on the recommended values,
        whenever possible, is strongly encouraged.

        * **major** = A major release may contain significant changes or may introduce breaking changes.
        * **minor** = A minor release, also known as an update, may contain a smaller number of changes than major
            releases.
        * **patch** = Patch releases are typically unplanned and may resolve defects or important security issues.
        * **pre-release** = A pre-release may include alpha, beta, or release candidates and typically have limited
            support. They provide the ability to preview a release prior to its general availability.
        * **internal** = Internal releases are not for public consumption and are intended to be used exclusively by the
            project or manufacturer that produced it.
        """
        return self._type

    @type.setter
    def type(self, type_: str) -> None:
        self._type = type_

    @property
    def title(self) -> Optional[str]:
        """
        The title of the release.
        """
        return self._title

    @title.setter
    def title(self, title: Optional[str]) -> None:
        self._title = title

    @property
    def featured_image(self) -> Optional[XsUri]:
        """
        The URL to an image that may be prominently displayed with the release note.
        """
        return self._featured_image

    @featured_image.setter
    def featured_image(self, featured_image: Optional[XsUri]) -> None:
        self._featured_image = featured_image

    @property
    def social_image(self) -> Optional[XsUri]:
        """
        The URL to an image that may be used in messaging on social media platforms.
        """
        return self._social_image

    @social_image.setter
    def social_image(self, social_image: Optional[XsUri]) -> None:
        self._social_image = social_image

    @property
    def description(self) -> Optional[str]:
        """
        A short description of the release.
        """
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self._description = description

    @property
    def timestamp(self) -> Optional[datetime]:
        """
        The date and time (timestamp) when the release note was created.
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: Optional[datetime]) -> None:
        self._timestamp = timestamp

    @property
    def aliases(self) -> "SortedSet[str]":
        """
        One or more alternate names the release may be referred to. This may include unofficial terms used by
        development and marketing teams (e.g. code names).

        Returns:
            Set of `str`
        """
        return self._aliases

    @aliases.setter
    def aliases(self, aliases: Iterable[str]) -> None:
        self._aliases = SortedSet(aliases)

    @property
    def tags(self) -> "SortedSet[str]":
        """
        One or more tags that may aid in search or retrieval of the release note.

        Returns:
            Set of `str`
        """
        return self._tags

    @tags.setter
    def tags(self, tags: Iterable[str]) -> None:
        self._tags = SortedSet(tags)

    @property
    def resolves(self) -> "SortedSet[IssueType]":
        """
        A collection of issues that have been resolved.

        Returns:
            Set of `IssueType`
        """
        return self._resolves

    @resolves.setter
    def resolves(self, resolves: Iterable[IssueType]) -> None:
        self._resolves = SortedSet(resolves)

    @property
    def notes(self) -> "SortedSet[Note]":
        """
        Zero or more release notes containing the locale and content. Multiple note elements may be specified to support
        release notes in a wide variety of languages.

        Returns:
            Set of `Note`
        """
        return self._notes

    @notes.setter
    def notes(self, notes: Iterable[Note]) -> None:
        self._notes = SortedSet(notes)

    @property
    def properties(self) -> "SortedSet[Property]":
        """
        Provides the ability to document properties in a name-value store. This provides flexibility to include data not
        officially supported in the standard without having to use additional namespaces or create extensions. Unlike
        key-value stores, properties support duplicate names, each potentially having different values.

        Returns:
            Set of `Property`
        """
        return self._properties

    @properties.setter
    def properties(self, properties: Iterable[Property]) -> None:
        self._properties = SortedSet(properties)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ReleaseNotes):
            return hash(other) == hash(self)
        return False

    def __hash__(self) -> int:
        return hash((
            self.type, self.title, self.featured_image, self.social_image, self.description, self.timestamp,
            tuple(self.aliases), tuple(self.tags), tuple(self.resolves), tuple(self.notes), tuple(self.properties)
        ))

    def __repr__(self) -> str:
        return f'<ReleaseNotes type={self.type}, title={self.title}>'
